import numpy as np

from trajectorize._c_extension import ffi, lib


def process_sva_buffer(sva_struct, length: int) -> np.ndarray:
    # Process memory buffer
    # Format: [x, y, z, vx, vy, vz, t] x n
    arr = np.frombuffer(ffi.buffer(sva_struct.mem_buffer, 7 *
                                   8 * length), dtype=np.float64)
    arr.shape = (length, 7)
    copied_arr = np.copy(arr)

    # Free memory
    lib.free_StateVectorArray(sva_struct)

    return copied_arr



import numpy as np
import pytest

from trajectorize.orbit import conic_kepler
from trajectorize.tester_utils.fake_body import EARTH_SI


def test_orbital_period():
    # Create an example orbit of known period
    # Circular orbit with a = 7378.1 km
    # T = 01:45:07.12 (hh:mm:ss) = 6372.12 s

    elements = conic_kepler.KeplerianElements(
        semi_major_axis=7378.14e3,
        eccentricity=0,
        inclination=0,
        longitude_of_ascending_node=0,
        argument_of_periapsis=0,
        true_anomaly=0,
        epoch=0
    )

    orbit = conic_kepler.KeplerianOrbit(elements, EARTH_SI)

    # Assert that the orbital period is correct
    assert orbit.T == pytest.approx(6307.12, abs=1e-2)


def test_kepler_equation_solver():
    # From Curtis Orbital Mechanics Appendix D
    M = 3.6029
    e = 0.37255
    result = conic_kepler.solve_kepler_equation(M, e)
    assert result == pytest.approx(3.479422)


def test_ke_from_state_vector():
    # Using Curtis example 4.3
    position = np.array([-6045, -3490, 2500])*1e3
    velocity = np.array([-3.457, 6.618, 2.533])*1e3

    orbit = conic_kepler.KeplerianOrbit.from_state_vector(position,
                                                          velocity,
                                                          0,
                                                          EARTH_SI)

    assert orbit.ke.semi_major_axis == pytest.approx(8788e3, rel=1e-3)
    assert orbit.ke.eccentricity == pytest.approx(0.1712, rel=1e-4)
    assert orbit.ke.inclination == pytest.approx(np.deg2rad(153.2), rel=1e-3)
    assert orbit.ke.longitude_of_ascending_node == pytest.approx(
        np.deg2rad(255.3), rel=1e-4)
    assert orbit.ke.argument_of_periapsis == pytest.approx(
        np.deg2rad(20.07), rel=1e-4)
    assert orbit.ke.true_anomaly == pytest.approx(np.deg2rad(28.45), rel=1e-3)


def test_state_vector_from_ke():
    # Using Curtis example 4.3
    position = np.array([-6045, -3490, 2500])*1e3
    velocity = np.array([-3.457, 6.618, 2.533])*1e3

    orbit = conic_kepler.KeplerianOrbit.from_state_vector(position,
                                                          velocity,
                                                          0,
                                                          EARTH_SI)

    sv = orbit.state_vector
    assert np.allclose(sv.position, position)
    assert np.allclose(sv.velocity, velocity)



from matplotlib import pyplot as plt

from trajectorize.ephemeris.kerbol_system import Body
from trajectorize.visualizers.display_utils import display_or_save_plot
from trajectorize.visualizers.porkchop_plot import porkchop_plot_ejection

if __name__ == "__main__":
    kerbin = Body.from_name("Kerbin")
    duna = Body.from_name("Duna")

    t1_min = 0
    t1_max = 852 * 86400 / 4

    tof_min = 151 * 86400 / 4
    tof_max = 453 * 86400 / 4

    plt.style.use('dark_background')

    fig, ax = plt.subplots(figsize=(8, 7))
    porkchop_plot_ejection(kerbin, duna, ax, t1_min,
                           t1_max, tof_min, tof_max, 70000)

    display_or_save_plot("kerbin_duna_porkchop.png")



import numpy as np

from trajectorize.trajectory.transfer_orbit import (TrajectoryDirection,
                                                    solve_lambert_problem)


def test_orbit_determination():
    # Test case from Curtis for Earth orbit determination
    r1 = np.array([5000, 10000, 2100])
    r2 = np.array([-14600, 2500, 7000])

    dt = 3600
    direction = TrajectoryDirection.PROGRADE
    mu = 398600  # km^3/s^2; Earth

    v1, v2 = solve_lambert_problem(r1, r2, dt, mu, direction)

    assert np.allclose(v1, np.array([-5.99249, 1.92536, 3.24564]), atol=1e-3)
    assert np.allclose(v2, np.array([-3.31246, -4.19662, -0.385288]),
                       atol=1e-3)



import pytest

from trajectorize.ephemeris.kerbol_system import (Body, BodyEnum,
                                                  state_vector_at_time)


def test_kerbin_parameters():
    kerbin = Body.from_name("Kerbin")

    assert kerbin.mass == pytest.approx(5.2915158e22)
    assert kerbin.radius == pytest.approx(600000)


def test_ephemeris_1():
    initial_kerbin_State = state_vector_at_time(0, BodyEnum.KERBOL,
                                                BodyEnum.KERBIN)

    assert initial_kerbin_State.time == 0
    print(initial_kerbin_State)



from argparse import ArgumentParser

import numpy as np
from matplotlib import pyplot as plt
from matplotlib.animation import FuncAnimation, PillowWriter

from trajectorize.ephemeris.kerbol_system import (Body, BodyEnum,
                                                  state_vector_at_time)

if __name__ == "__main__":

    parser = ArgumentParser(description="Kerbol System Animation")

    '''
    Command Line Options

    --save: Save the animation to a file
    --length: Length of the animation in years (each year is 426.08 Kerbin days)
    '''
    parser.add_argument("--save", action="store_true",
                        help="Save the animation to a file")
    parser.add_argument(
        "--length",
        type=int,
        default=10,
        help="Length of the animation in years (each year is 426.08 Kerbin days)")

    args = parser.parse_args()

    years = args.length
    save_anim = args.save

    KERBIN_DAY = 21600
    KERBIN_YEAR = 426.08 * KERBIN_DAY

    # Generate animated frames
    t_ut = np.linspace(0, KERBIN_YEAR * years, 500)

    plt.style.use('dark_background')

    figsize = (3.5, 3.5) if save_anim else (8, 7)

    fig, ax = plt.subplots(figsize=figsize)

    planets = Body.planets()

    # place title inside the axes so it can blit
    title = ax.set_title(
        f"Kerbol System, UT = {round(t_ut[0]/KERBIN_DAY)}d",
        y=0.8, x=0.5, fontsize=10)

    ax.set_axis_off()
    ax.set_aspect('equal')
    fig.tight_layout()

    # Plot Kerbol and orbits of planets (static; no need to redraw)
    kerbol = Body.from_name("Kerbol")
    ax.plot(0, 0, 'o', markersize=5, color=kerbol.colour_hex)

    for body in planets:
        locus = body.orbit_locus(100)
        ax.plot(locus[:, 0], locus[:, 1],
                color=body.colour_hex,
                label=body.name)

    # Construct dynamic artists; these are the planets and their markers
    planet_ephemerides = [[None] * len(planets) for _ in range(len(t_ut))]

    for i, ut in enumerate(t_ut):
        for j, b in enumerate(planets):
            planet_ephemerides[i][j] = state_vector_at_time(ut,
                                                            BodyEnum.KERBOL,
                                                            b.body_id)

    planet_artists = [None for _ in range(len(planets))]

    for i, b in enumerate(planets):
        marker, = ax.plot(planet_ephemerides[0][i].position[0],
                          planet_ephemerides[0][i].position[1],
                          'o', markersize=5, color=b.colour_hex)
        planet_artists[i] = marker

    # Construct animation

    def animate(i):
        for j, b in enumerate(planets):
            planet_artists[j].set_data(planet_ephemerides[i][j].position[0],
                                       planet_ephemerides[i][j].position[1])

        title.set_text(f"Kerbol System, UT = {round(t_ut[i]/KERBIN_DAY)}d")

        return planet_artists + [title]

    animation = FuncAnimation(fig, animate, frames=len(
        t_ut) - 1, interval=30, blit=True)

    if plt.get_backend() == "agg":
        print("Non-GUI backend detected. Cannot show animation.")
    else:
        plt.show()

    if save_anim:
        # Write animation to file
        # TODO: optimize GIF to reduce file size
        writer = PillowWriter(fps=30,
                              metadata=dict(artist='Me'),
                              bitrate=50)
        animation.save('kerbol_system.gif', writer=writer)



from matplotlib.artist import Artist
from matplotlib.axes import Axes

from trajectorize.ephemeris.kerbol_system import (Body, BodyEnum,
                                                  state_vector_at_time)
from trajectorize.orbit.conic_kepler import KeplerianElements, KeplerianOrbit


def plot_body_rel_kerbol(body: Body, ut: float, ax: Axes,
                         plot_orbit: bool = True,
                         num_ellipse_samples: int = 1000,
                         marker_str: str = "o",
                         markersize: int = 5) -> "tuple(Artist)":
    '''
    Plots a celestial body on a matplotlib axes object.

    Plot centered around the Kerbol system origin.

    Parameters
    ----------
    body: CelestialBody
        The body to plot.
    ut: float
        Universal time in seconds. (Game Time)
    ax: Axes
        The axes on which to plot the body.
    mu: float
        The gravitational parameter of the central body.
    plot_orbit: bool
        Whether to plot the orbit of the body.
    num_ellipse_samples: int
        The number of samples to use when plotting the orbit ellipse.
    marker_str: str
        The marker to use when plotting the body. (see matplotlib docs)
    markersize: int
        The size of the marker to use when plotting the body.
    '''
    obj_state_vec = state_vector_at_time(ut, BodyEnum.KERBOL,
                                         body.body_id)
    marker, = ax.plot(obj_state_vec.position[0], obj_state_vec.position[1],
                      marker=marker_str, markersize=markersize,
                      color=body.colour_hex,
                      label=body.name)

    if plot_orbit:
        obj_locus = body.orbit_locus(num_ellipse_samples)
        path, = ax.plot(obj_locus[:, 0], obj_locus[:, 1],
                        color=body.colour_hex)
        return marker, path
    else:
        return marker


def kerbol_system_plot(ut: float, ax: Axes,
                       num_ellipse_samples: int = 1000,
                       show_legend: bool = True,
                       markersize: int = True) -> "tuple(Artist)":
    '''
    Higher level "prefab" function to generate a pretty looking plot
    of the Kerbol system.
    '''
    planets = Body.planets()
    for body in planets:
        plot_body_rel_kerbol(
            body, ut, ax, num_ellipse_samples=num_ellipse_samples,
            markersize=markersize)

    # Plot Kerbol
    kerbol = Body.from_name("Kerbol")
    ax.plot(0, 0, 'o', markersize=markersize, color=kerbol.colour_hex)

    # Styling
    ax.set_axis_off()
    ax.set_aspect("equal")

    ax.set_title(f"Kerbol System at UT = {ut} seconds")

    if show_legend:
        ax.legend()



from enum import IntEnum

import numpy as np

from trajectorize._c_extension import lib
from trajectorize.ephemeris.kerbol_system import Body
from trajectorize.math_lib.math_interfaces import (np_array_from_vec3,
                                                   vec3_from_np_array)
from trajectorize.orbit.conic_kepler import KeplerianElements, KeplerianOrbit


class TrajectoryDirection(IntEnum):
    PROGRADE = lib.PROGRADE
    RETROGRADE = lib.RETROGRADE


def planetary_transfer_orbit(body1: Body, body2: Body, t1: float, t2: float) \
        -> KeplerianOrbit:
    parent = Body.from_identifier(body1.parent_id)
    sol = lib.planetary_transfer_orbit(body1.c_data, body2.c_data, t1, t2)

    ke = KeplerianElements.from_c_data(sol)
    orbit = KeplerianOrbit(ke, parent)
    return orbit


def get_excess_velocity(body: Body, orbit: KeplerianOrbit, time: float) \
        -> np.ndarray:
    '''
    Gets the excess velocity between the transfer orbit and a celestial
    body at some time.
    Used to patch the transfer orbit into the body orbit.
    '''
    return np_array_from_vec3(
        lib.excess_velocity_at_body(body.c_data, orbit.ke.c_data, time))


def estimate_delta_v(body: Body, excess_velocity: float,
                     periapsis_radius: float) -> float:
    '''
    Estimates the delta-v required to patch the transfer orbit into the body orbit.
    '''
    return lib.delta_v_req(body.c_data, excess_velocity, periapsis_radius)


def solve_lambert_problem(r1: np.ndarray, r2: np.ndarray,
                          dt: float, mu: float,
                          direction: TrajectoryDirection) \
        -> "tuple(np.ndarray, np.ndarray)":
    r1_c = vec3_from_np_array(r1)
    r2_c = vec3_from_np_array(r2)
    sol = lib.lambert(r1_c, r2_c, dt, mu, int(direction))
    v1 = np_array_from_vec3(sol.v1)
    v2 = np_array_from_vec3(sol.v2)
    return v1, v2


def trajectory_ejection_dv(t1: float, t2: float,
                           body1: Body, body2: Body,
                           parking_orbit_alt: float = 100000):
    '''
    Delta-v required to eject from body1 and enter a transfer orbit to body2.

    Parameters
    ----------
    t1: float
        Time of departure from body1 (universal time, seconds)
    t2: float
        Time of arrival at body2 (universal time, seconds)
    body1: Body
        Departure body
    body2: Body
        Arrival body
    parking_orbit_alt: float
        Altitude of parking orbit (meters) above body1's surface.
        e.g. for Kerbin, it's about 70 km for an ideal orbit just outside
        the atmosphere.
    '''
    try:
        orbit = planetary_transfer_orbit(
            body1, body2, t1, t2)
        excess_velocity = get_excess_velocity(body1, orbit, t1)
        excess_speed = np.linalg.norm(excess_velocity)
        delta_v = estimate_delta_v(
            body1, excess_speed,
            orbit.ke.semi_major_axis + parking_orbit_alt)
        return delta_v
    except Exception:
        return np.nan


# standalone test
if __name__ == "__main__":
    # Test transfer from Kerbin to Duna
    kerbin = Body.from_name("Kerbin")
    duna = Body.from_name("Duna")

    t1 = 5091552
    t2 = 10679760

    transfer_orbit = planetary_transfer_orbit(kerbin, duna, t1, t2)
    kerbin_excess_velocity = get_excess_velocity(kerbin, transfer_orbit, t1)

    v_inf_kerbin = np.linalg.norm(kerbin_excess_velocity)
    v_inf_duna = np.linalg.norm(get_excess_velocity(duna, transfer_orbit, t2))
    r_pe_kerbin = kerbin.radius + 70000  # 70 km parking orbit
    dv_est_kerbin = estimate_delta_v(kerbin, v_inf_kerbin, r_pe_kerbin)
    r_pe_duna = duna.radius + 50000  # 50 km target orbit
    dv_est_duna = estimate_delta_v(duna, v_inf_duna, r_pe_duna)

    print(f"Estimated ejection delta-v: {dv_est_kerbin:.2f} m/s")
    print(f"C3 wrt Kerbin: {v_inf_kerbin**2/1e6:.2f} km^2/s^2")
    print(f"Estimate capture delta-v: {dv_est_duna:.2f} m/s")



import numpy as np
import pytest

from trajectorize.orbit.universal_kepler import UniversalKeplerOrbit


# Assure that propagation is time invariant and is not affected by start time
@pytest.mark.parametrize("t_offset", [0, -100, 100])
def test_universal_kepler(t_offset):
    r0 = np.array([7000e3, -12124e3, 0])
    v0 = np.array([2.6679e3, 4.6210e3, 0])
    t = 3600 + t_offset

    orb = UniversalKeplerOrbit(r0, v0, t_offset, 3.986e14)

    orb2 = orb.propagate(t)

    assert np.allclose(orb2.position, np.array(
        [-3297.7686252e3, 7413.39664579e3, 0]))
    assert np.allclose(orb2.velocity, np.array(
        [-8.29760302e3, -0.96404494e3, -0.]))


@pytest.mark.parametrize("t_offset", [0, -100, 100])
def test_universal_kepler_vec(t_offset):
    r0 = np.array([7000e3, -12124e3, 0])
    v0 = np.array([2.6679e3, 4.6210e3, 0])
    t = np.array([1800, 3600, 5400, 7200]) + t_offset

    orb = UniversalKeplerOrbit(r0, v0, t_offset, 3.986e14)

    orb2 = orb.propagate_vec(t)

    assert np.allclose(orb2.position[1, :], np.array(
        [-3297.7686252e3, 7413.39664579e3, 0]))
    assert np.allclose(orb2.velocity[1, :], np.array(
        [-8.29760302e3, -0.96404494e3, -0.]))



class FakeBody:
    '''
    Defines a body with a monkey-patched mu attribute which behaves like a
    Body from the kerbol_system module.
    '''

    def __init__(self, mu):
        self.mu = mu


EARTH_SI = FakeBody(3.986004418e14)
EARTH_KM = FakeBody(3.986004418e5)



# Wrapper for C code

from dataclasses import dataclass

import numpy as np

from trajectorize._c_extension import ffi, lib
from trajectorize.c_ext_utils.process_sva_buffer import process_sva_buffer
from trajectorize.math_lib.math_interfaces import (np_array_from_vec3,
                                                   vec3_from_np_array)


@dataclass
class UniversalKeplerOrbit:
    position: np.ndarray       # Either a single position or list of positions
    velocity: np.ndarray       # Either a single velocity or list of velocities
    time: "float|np.ndarray"   # used when called vectorized
    mu: float

    def propagate(self, t: float) -> "UniversalKeplerOrbit":
        orbit_c = ffi.new("struct StateVector *", {
            "position": vec3_from_np_array(self.position),
            "velocity": vec3_from_np_array(self.velocity),
            "time": self.time
        })[0]

        new_orbit = lib.state_vec_orbit_prop(t, orbit_c, self.mu)

        new_position = np_array_from_vec3(new_orbit.position)
        new_velocity = np_array_from_vec3(new_orbit.velocity)

        return UniversalKeplerOrbit(new_position, new_velocity,
                                    self.time + t, self.mu)

    def propagate_vec(self, times: np.ndarray) -> "UniversalKeplerOrbit":
        orbit_c = ffi.new("struct StateVector *", {
            "position": vec3_from_np_array(self.position),
            "velocity": vec3_from_np_array(self.velocity),
            "time": self.time
        })[0]

        c_times = ffi.new("double[]", times.tolist())
        prop_orbit = lib.state_vec_orbit_prop_many(
            len(times), c_times, orbit_c, self.mu)

        arr = process_sva_buffer(prop_orbit, len(times))

        positions = arr[:, :3]
        velocities = arr[:, 3:6]

        return UniversalKeplerOrbit(positions, velocities,
                                    times, self.mu)



from matplotlib import pyplot as plt

from trajectorize.visualizers.celestial_body_plot import kerbol_system_plot
from trajectorize.visualizers.display_utils import display_or_save_plot

if __name__ == "__main__":

    plt.style.use('dark_background')

    fig, ax = plt.subplots(figsize=(8, 7))

    kerbol_system_plot(0, ax)
    display_or_save_plot("kerbol_system.png")



# Wrapper for C code
from dataclasses import dataclass
from enum import IntEnum

import numpy as np

from trajectorize._c_extension import ffi, lib
from trajectorize.ephemeris.state_vector import StateVector


class BodyEnum(IntEnum):
    KERBOL = lib.KERBOL
    MOHO = lib.MOHO
    EVE = lib.EVE
    GILLY = lib.GILLY
    KERBIN = lib.KERBIN
    MUN = lib.MUN
    MINMUS = lib.MINMUS
    DUNA = lib.DUNA
    IKE = lib.IKE
    DRES = lib.DRES
    JOOL = lib.JOOL
    LAYTHE = lib.LAYTHE
    VALL = lib.VALL
    TYLO = lib.TYLO
    BOP = lib.BOP
    POL = lib.POL
    EELOO = lib.EELOO

# PlanetaryKeplerianElements struct class


@dataclass
class PlanetaryKeplerianElements:
    '''
    Equivalent of C PlanetaryKeplerianElements struct.
    '''
    semi_major_axis: float
    eccentricity: float
    inclination: float
    longitude_of_ascending_node: float
    argument_of_periapsis: float
    mean_anomaly_at_epoch: float

    @classmethod
    def from_c_data(cls, cdata):
        return cls(cdata.semi_major_axis,
                   cdata.eccentricity,
                   cdata.inclination,
                   cdata.longitude_of_ascending_node,
                   cdata.argument_of_periapsis,
                   cdata.mean_anomaly_at_epoch)

    @property
    def c_data(self):
        '''
        Returns C struct compatible with library functions.
        '''
        return ffi.new('PlanetaryKeplerianElements *', self.__dict__)[0]


# Planets

@ dataclass
class Body:
    body_id: BodyEnum
    parent_id: BodyEnum
    mass: float
    mu: float
    radius: float
    atmosphere_height: float
    orbit: PlanetaryKeplerianElements
    soi_radius: float
    colour: int

    @ property
    def name(self) -> str:
        return BodyEnum(self.body_id).name.title()

    @ property
    def colour_hex(self) -> str:
        # returns as hex string
        return f"#{self.colour:06x}"

    @classmethod
    def from_identifier(cls, identifier: "BodyEnum|int") -> "Body":
        '''
        Create a Body object from an identifier
        '''
        if identifier not in BodyEnum:
            raise ValueError("Invalid planet")

        body = lib.kerbol_system_bodies[int(identifier)]

        return cls(BodyEnum(body.body_id),
                   BodyEnum(body.parent_id),
                   body.mass,
                   body.mu,
                   body.radius,
                   body.atmosphere_height,
                   PlanetaryKeplerianElements.from_c_data(body.orbit),
                   body.soi_radius,
                   body.colour)

    @ classmethod
    def from_name(cls, name: str) -> "Body":
        '''
        Create a Body object from a string
        '''
        name_dict = {k.casefold(): v for k, v in BodyEnum.__members__.items()}

        if name.casefold() not in name_dict:
            raise ValueError("Invalid planet")

        return cls.from_identifier(name_dict[name.casefold()])

    @ property
    def c_data(self):
        '''
        Returns C struct compatible with library functions.
        '''
        return lib.kerbol_system_bodies[self.body_id.value]

    def orbit_locus(self, num_points: int = 1000) -> np.ndarray:
        '''
        Return a (N, 3) array of points on the orbit of the body
        in state space relative to its parent.
        '''
        parent = Body.from_identifier(self.parent_id)
        state_vec_arr = lib.pke_state_locus(self.orbit.c_data, parent.mu,
                                            num_points)

        # process memory buffer
        # Format: [x, y, z, vx, vy, vz, t] x n
        arr = np.frombuffer(ffi.buffer(state_vec_arr.mem_buffer, 7 *
                                       8 * num_points), dtype=np.float64)
        arr.shape = (num_points, 7)
        positions = np.copy(arr[:, :3])  # copy to avoid memory leak

        lib.free_StateVectorArray(state_vec_arr)

        return positions

    @property
    def parent(self) -> "Body":
        '''
        Returns the parent body of this body.
        '''
        return Body.from_identifier(self.parent_id)

    @classmethod
    def all_bodies(cls) -> "list[Body]":
        '''
        Returns a list of all bodies in the system.
        '''
        return [cls.from_identifier(i) for i in BodyEnum]

    @classmethod
    def planets(cls) -> "list[Body]":
        '''
        Returns a list of all planets in the system
        (i.e. Kerbol is their parent)
        '''
        bodies = cls.all_bodies()
        return [body for body in bodies if body.parent_id == BodyEnum.KERBOL
                and body.body_id != BodyEnum.KERBOL]


def state_vector_at_time(t: float, parent: BodyEnum,
                         child: BodyEnum) -> StateVector:
    cdata = lib.get_rel_state_at_time(t, parent.value, child.value)
    return StateVector.from_c_data(cdata)



from functools import partial
from multiprocessing import Pool, cpu_count

import numpy as np
from matplotlib.artist import Artist
from matplotlib.axes import Axes
from tqdm import tqdm

from trajectorize.ephemeris.kerbol_system import Body
from trajectorize.trajectory.transfer_orbit import trajectory_ejection_dv
from trajectorize.visualizers.parula_colourmap import parula_map


def porkchop_plot_ejection(body1: Body, body2: Body, ax: Axes,
                           t1_min: float, t1_max: float,
                           tof_min: float, tof_max: float,
                           parking_orbit_alt: float) -> Artist:
    '''
    Plots a porkchop plot of the ejection times from body1 to body2.
    '''

    t1 = np.linspace(t1_min, t1_max, 200)

    tof = np.linspace(tof_min, tof_max, 200)

    g_t1, g_tof = np.meshgrid(t1, tof)

    # construct iterable of arguments for the pool
    t1t2_vals = zip(g_t1.ravel(), g_t1.ravel() + g_tof.ravel())

    # make a partial function to pass to the pool for a given t1 and t2
    # using star map to unpack the arguments

    with Pool(cpu_count() - 1) as pool:
        dvs = pool.starmap(partial(trajectory_ejection_dv,
                                   body1=body1,
                                   body2=body2,
                                   parking_orbit_alt=parking_orbit_alt),
                           tqdm(t1t2_vals, total=len(t1) * len(tof),
                                desc="Calculating Porkchop Plot"))

    # convert the time values to kerbin days

    t1_kerbin_days = g_t1 / 86400 * 4
    tof_kerbin_days = g_tof / 86400 * 4
    # reshape dv to match t1_kerbin_days and tof_kerbin_days
    dv = np.array(dvs).reshape(t1_kerbin_days.shape)

    # plot the porkchop plot
    mesh = ax.pcolormesh(t1_kerbin_days, tof_kerbin_days, dv, cmap="jet",
                         shading='gouraud', vmin=min(dvs),
                         vmax=np.percentile(dvs, 99))

    # add colorbar
    tick_marks = np.linspace(np.percentile(
        dvs, 0.5), np.percentile(dvs, 99), 10)
    colorbar = ax.figure.colorbar(mesh, ax=ax, fraction=0.05, extend='max')
    colorbar.set_ticks(tick_marks)

    # show min dv point using lines
    min_dv_idx = np.unravel_index(np.argmin(dv), dv.shape)
    t1_min_dv = t1_kerbin_days[min_dv_idx]
    tof_min_dv = tof_kerbin_days[min_dv_idx]
    ax.axvline(t1_min_dv, color='w', linestyle='--')
    ax.axhline(tof_min_dv, color='w', linestyle='--')

    # set the labels
    ax.set_title(f'Porkchop Plot from {body1.name} to {body2.name}')
    ax.set_xlabel('Departure Time (Kerbin Days)')
    ax.set_ylabel('Time of Flight (Kerbin Days)')

    return (mesh, colorbar)



from dataclasses import dataclass

import numpy as np

from trajectorize.math_lib.math_interfaces import np_array_from_vec3


@dataclass
class StateVector:
    position: np.ndarray
    velocity: np.ndarray
    time: float

    @classmethod
    def from_c_data(cls: "StateVector", cdata):
        return cls(np_array_from_vec3(cdata.position),
                   np_array_from_vec3(cdata.velocity),
                   time=cdata.time)



from matplotlib import pyplot as plt


def display_or_save_plot(filename: str):
    # Determine backend; if GUI, show plot, else save to file

    # Agg backend is used for non-GUI environments
    if plt.get_backend() == "agg":
        plt.savefig(filename)
        print(f"Non-GUI backend detected. Plot saved to {filename}")
    else:
        plt.show()



import sys

from cffi import FFI

include_dir = "include" if __name__ != "__main__" else "../include"
src_path = "src/trajectorize" if __name__ != "__main__" else "trajectorize"


def read_and_cleanse_header(filename: str) -> str:
    '''
    Reads a header file (.h) and removes macros,
    returning the contents as a string.
    '''
    filepath = f"{include_dir}/{filename}"

    with open(str(filepath), "r", encoding="utf-8") as f:
        lines = f.readlines()
        clean_lines = [line for line in lines if not line.startswith("#")]
        return ''.join(clean_lines)


def read_and_cleanse_many_headers(*filenames: str) -> str:
    '''
    Utility function for reading and cleansing multiple header files,
    joining them into a single string suitable for cffi.cdef().
    '''
    return ''.join([read_and_cleanse_header(filename)
                   for filename in filenames])


ffi = FFI()

'''
Add files in dependency order.
Generally, the order is:
- "..._types.h"
- everything else that's called at the top level
- files "in the middle" need not be included e.g. math library functions

This is because header files further down the list may depend on types defined
upstream.
'''
ffi.cdef(read_and_cleanse_many_headers("vec_math_types.h",
                                       "state_vector_types.h",
                                       "keplerian_element_types.h",
                                       "kerbol_system_types.h",
                                       "kerbol_system_bodies.h",
                                       "kerbol_system_ephemeris.h",
                                       "conic_kepler.h",
                                       "universal_kepler.h",
                                       "lambert.h",
                                       "transfer_orbit.h",
                                       "delta_v_estimate.h"))

'''
Determine compiler flags and linker args depending on platform.

GCC options
-std=c99: Use C99 standard
-lm: Link math library
-lc: Link C standard library
-march=native: Optimize for the current CPU
-O3: Optimize for speed

MSVC options


Compiler type is determined by the value of sys.platform.
'''
if sys.platform == "linux" or sys.platform == "linux2":
    # linux
    extra_compile_args = ["-std=c99", "-lm", "-lc", "-O3"]
else:
    # MSVC is already c99 compliant, so no need to specify
    extra_compile_args = None

# Include only the top-level header files.
ffi.set_source("trajectorize._c_extension",
               '''
               #include "kerbol_system_bodies.h"
               #include "kerbol_system_ephemeris.h"
               #include "conic_kepler.h"
               #include "universal_kepler.h"
               #include "lambert.h"
               #include "transfer_orbit.h"
               #include "delta_v_estimate.h"
               ''',
               sources=[f"{src_path}/math_lib/orbit_math.c",
                        f"{src_path}/math_lib/rotations.c",
                        f"{src_path}/math_lib/stumpuff_functions.c",
                        f"{src_path}/ephemeris/kerbol_system_bodies.c",
                        f"{src_path}/ephemeris/kerbol_system_ephemeris.c",
                        f"{src_path}/orbit/conic_kepler.c",
                        f"{src_path}/orbit/universal_kepler.c",
                        f"{src_path}/trajectory/lambert.c",
                        f"{src_path}/trajectory/transfer_orbit.c",
                        f"{src_path}/trajectory/delta_v_estimate.c"],
               include_dirs=[include_dir],
               extra_compile_args=extra_compile_args)

if __name__ == "__main__":
    # For debug building
    ffi.compile(verbose=True)



import numpy as np

from trajectorize._c_extension import ffi


def vec3_from_np_array(np_array: np.ndarray) -> ffi.CData:
    return ffi.new("Vector3*", {"v": np_array.tolist()})[0]


def np_array_from_vec3(vec3: ffi.CData) -> np.ndarray:
    return np.copy(np.frombuffer(ffi.buffer(vec3.v, 3 * 8), dtype=np.float64))



#!/usr/bin/env python

import pathlib

from setuptools import setup

if __name__ == "__main__":
    # Make a relative path for the cffi module
    cffi_dir = pathlib.Path(__file__).parent / "src" / \
        "trajectorize" / "c_ext_utils"
    setup(cffi_modules=[str(cffi_dir / "build_c_extension.py:ffi")])



from time import perf_counter

import matplotlib.pyplot as plt
import numpy as np

from trajectorize.orbit.universal_kepler import UniversalKeplerOrbit
from trajectorize.visualizers.display_utils import display_or_save_plot

if __name__ == "__main__":
    r0 = np.array([7000, -12124, 0])
    v0 = np.array([2.6679, 4.6210, 0])

    demo_orbit = UniversalKeplerOrbit(r0, v0, 0, 398600)

    t = np.linspace(0, 34000, 1000)

    start = perf_counter()

    orb = demo_orbit.propagate_vec(t)

    r = orb.position
    v = orb.velocity

    plt.style.use("dark_background")
    fig = plt.figure(figsize=(12, 5), tight_layout=True)

    earth = [plt.Circle((0, 0), 6378, color='#00a7a5') for _ in range(3)]
    axes = []

    for i in range(3):
        axes.append(fig.add_subplot(231 + i))
        axes[i].add_patch(earth[i])
        axes[i].set_aspect("equal")

    x = r[:, 0]
    y = r[:, 1]
    z = r[:, 2]
    axes[0].plot(x, y)
    axes[0].set_ylabel("y position (km)")
    axes[0].set_xlabel("x position (km)")
    axes[1].plot(y, z)
    axes[1].set_ylabel("z position (km)")
    axes[1].set_xlabel("y position (km)")
    axes[2].plot(x, z)
    axes[2].set_ylabel("z position (km)")
    axes[2].set_xlabel("x position (km)")
    axes[0].set_title("xy")
    axes[1].set_title("yz")
    axes[2].set_title("xz")

    axes.append(fig.add_subplot(212))
    # Plot altitude above each
    altitude = np.sqrt(x**2 + y**2 + z**2) - 6378
    axes[3].plot(t, altitude)
    axes[3].set_ylabel("Altitude (km)")
    axes[3].set_xlabel("Mission Time (s)")

    fig.suptitle("Universal Keplerian Elements Demo")
    display_or_save_plot("orbit.png")



from matplotlib import pyplot as plt
from matplotlib.lines import Line2D

from trajectorize.ephemeris.kerbol_system import Body
from trajectorize.visualizers.display_utils import display_or_save_plot
from trajectorize.visualizers.planetary_transfer_plot import \
    plot_interplanetary_transfer

if __name__ == "__main__":
    kerbin = Body.from_name("Kerbin")
    duna = Body.from_name("Duna")

    t1 = 5091552
    t2 = 10679760

    plt.style.use('dark_background')

    fig, ax = plt.subplots(figsize=(8, 7))

    plot_interplanetary_transfer(kerbin, duna, t1, t2, ax)

    # Plot custom legend; label circle marker = t1, label diamond marker = t2
    legend_elements = [Line2D([0], [0], marker='o', markerfacecolor='w', lw=0,
                              label='Planet Positions at Kerbin Departure'),
                       Line2D([0], [0], marker='D', markerfacecolor='w', lw=0,
                              label='Planet Positions at Duna Arrival')]
    ax.legend(handles=legend_elements, loc='lower right')
    ax.set_title("Ballistic Transfer from Kerbin to Duna\n"
                 f"Departure: UT = {t1} seconds\n"
                 f"Time of Flight: {t2 - t1} seconds\n")
    fig.tight_layout()
    display_or_save_plot("kerbin_duna_transfer.png")



# Wrapper for C code
from dataclasses import dataclass

import numpy as np

from trajectorize._c_extension import ffi, lib
from trajectorize.c_ext_utils.process_sva_buffer import process_sva_buffer
from trajectorize.ephemeris.kerbol_system import Body
from trajectorize.ephemeris.state_vector import StateVector
from trajectorize.math_lib.math_interfaces import vec3_from_np_array


@dataclass
class KeplerianElements:
    '''
    Equivalent of C KeplerianElements struct.
    '''
    semi_major_axis: float
    eccentricity: float
    inclination: float
    longitude_of_ascending_node: float
    argument_of_periapsis: float
    true_anomaly: float
    epoch: float

    @classmethod
    def from_celestial_body(cls, body: Body, ut: float):
        '''
        Creates a KeplerianElements object from a celestial body
        at some time.
        '''

        # Use library function: ke_from_pke
        pke_c = body.orbit.c_data

        ke_c = lib.ke_from_pke(pke_c, ut, body.mu)
        return cls.from_c_data(ke_c)

    @classmethod
    def from_c_data(cls, cdata):
        return cls(cdata.semi_major_axis,
                   cdata.eccentricity,
                   cdata.inclination,
                   cdata.longitude_of_ascending_node,
                   cdata.argument_of_periapsis,
                   cdata.true_anomaly,
                   cdata.epoch)

    @property
    def c_data(self):
        '''
        Returns C struct compatible with library functions.
        '''
        return ffi.new('struct KeplerianElements *', self.__dict__)[0]


@ dataclass
class KeplerianOrbit:
    '''
    Convenience class storing a KeplerianElements object and a Body object.

    Used to perform composite calculations that require both Keplerian
    Elements as well as a central body's properties (mass, radius, etc.)
    '''
    ke: KeplerianElements
    parent_body: Body

    def get_locus(self, n: int) -> np.ndarray:
        '''
        Returns the locus of the orbit in state space.
        '''
        state_vec_arr = lib.ke_state_locus(
            self.ke.c_data, self.parent_body.mu, n)
        arr = process_sva_buffer(state_vec_arr, n)

        return arr[:, 0:3]

    @ property
    def T(self) -> float:
        '''
        Orbital period
        '''
        return lib.orbital_period(self.ke.semi_major_axis, self.parent_body.mu)

    @ classmethod
    def from_state_vector(cls, position: np.ndarray, velocity: np.ndarray,
                          epoch: float, parent_body: Body) -> "KeplerianOrbit":
        '''
        Returns the orbit from a state vector.
        '''
        state_vec = ffi.new('struct StateVector *', {
            'position': vec3_from_np_array(position),
            'velocity': vec3_from_np_array(velocity),
            'time': epoch
        })[0]

        ke = KeplerianElements.from_c_data(
            lib.ke_from_state_vector(state_vec, parent_body.mu))

        return cls(ke, parent_body)

    @property
    def state_vector(self):
        sv_c = lib.state_vector_from_ke(self.ke.c_data, self.parent_body.mu)
        return StateVector.from_c_data(sv_c)

    def propagate(self, t: float) -> "KeplerianOrbit":
        sol = lib.ke_orbit_prop(t, self.ke.c_data, self.parent_body.mu)

        ke = KeplerianElements.from_c_data(sol.ke)
        return KeplerianOrbit(ke, self.parent_body)

    def propagate_vec(self, times: np.ndarray) -> np.ndarray:
        c_times = ffi.new("double[]", times.tolist())
        prop_orbit = lib.ke_orbit_prop_many(len(times),
                                            c_times,
                                            self.ke.c_data,
                                            self.parent_body.mu)

        arr = process_sva_buffer(prop_orbit, len(times))
        return arr[:, 0:3]


def solve_kepler_equation(M: float, e: float) -> float:
    result: float = lib.E_from_M(M, e)
    return result



import numpy as np
from matplotlib.artist import Artist
from matplotlib.axes import Axes

from trajectorize.ephemeris.kerbol_system import Body
from trajectorize.trajectory.transfer_orbit import planetary_transfer_orbit
from trajectorize.visualizers.celestial_body_plot import plot_body_rel_kerbol


def plot_transfer_orbit(body1: Body, body2: Body, t1: float, t2: float,
                        ax: Axes,
                        plot_full_orbit: bool = False) -> "tuple(Artist)":
    '''
    Plots a transfer orbit between two bodies on a matplotlib axes object.

    Parameters
    ----------
    body1: Body
        The first body.
    body2: Body
        The second body.
    t1: float
        The departure time in seconds.
    t2: float
        The arrival time in seconds.
    ax: Axes
        The axes on which to plot the transfer orbit.
    plot_full_orbit: bool
        Whether or not to plot the full ellipse of the orbit. Plots only
        the portion of the orbit between t1 and t2 if False.
    '''
    # Get transfer orbit
    transfer_orbit = planetary_transfer_orbit(body1, body2, t1, t2)

    if plot_full_orbit:
        transfer_locus = transfer_orbit.get_locus(500)
    else:
        eval_times = np.linspace(t1, t2, 500)
        transfer_locus = transfer_orbit.propagate_vec(eval_times)

    path, = ax.plot(transfer_locus[:, 0],
                    transfer_locus[:, 1], color="C1", lw=2)
    return path


def plot_interplanetary_transfer(body1: Body, body2: Body, t1: float,
                                 t2: float, ax: Axes) -> None:

    # Get transfer orbit
    plot_transfer_orbit(body1, body2, t1, t2, ax)

    # Plot planets
    plot_body_rel_kerbol(body1, t1, ax, plot_orbit=True,
                         marker_str="o", markersize=10)
    plot_body_rel_kerbol(body1, t2, ax, plot_orbit=False,
                         marker_str="D", markersize=10)
    plot_body_rel_kerbol(body2, t1, ax, plot_orbit=True,
                         marker_str="o", markersize=10)
    plot_body_rel_kerbol(body2, t2, ax, plot_orbit=False,
                         marker_str="D", markersize=10)
    plot_body_rel_kerbol(body1.parent, t1, ax, markersize=10)

    # Styling
    ax.set_axis_off()
    ax.set_aspect("equal")
